@@ -7,14 +7,50 @@ Make beautiful, interactive maps with Python and Leaflet.js
 
 '''
 
+from __future__ import absolute_import
 from __future__ import print_function
 from __future__ import division
+
 import codecs
+import functools
 import json
-import pandas as pd
+import os
+from uuid import uuid4
+
 from jinja2 import Environment, PackageLoader
 from pkg_resources import resource_string, resource_filename
-import utilities
+
+from folium import utilities
+from folium.six import text_type, binary_type, iteritems
+
+
+ENV = Environment(loader=PackageLoader('folium', 'templates'))
+
+
+def initialize_notebook():
+    """Initialize the IPython notebook display elements"""
+    try:
+        from IPython.core.display import display, HTML
+    except ImportError:
+        print("IPython Notebook could not be loaded.")
+
+    lib_css = ENV.get_template('ipynb_init_css.html')
+    lib_js = ENV.get_template('ipynb_init_js.html')
+    leaflet_dvf = ENV.get_template('leaflet-dvf.markers.min.js')
+
+    display(HTML(lib_css.render()))
+    display(HTML(lib_js.render({'leaflet_dvf': leaflet_dvf.render()})))
+
+def iter_obj(type):
+    '''Decorator to keep count of different map object types in self.mk_cnt'''
+    def decorator(func):
+        @functools.wraps(func)
+        def wrapper(self, *args, **kwargs):
+            self.mark_cnt[type] = self.mark_cnt.get(type, 0) + 1
+            func_result = func(self, *args, **kwargs)
+            return func_result
+        return wrapper
+    return decorator
 
 
 class Map(object):
@@ -29,6 +65,8 @@ class Map(object):
         tilesets or a custom tileset URL. The following tilesets are built-in
         to Folium. Pass any of the following to the "tiles" keyword:
             -"OpenStreetMap"
+            -"MapQuest Open"
+            -"MapQuest Open Aerial"
             -"Mapbox Bright" (Limited levels of zoom for free tiles)
             -"Mapbox Control Room" (Limited levels of zoom for free tiles)
             -"Stamen Terrain"
@@ -76,8 +114,11 @@ class Map(object):
 
         '''
 
-        #Init Map type
+        #Inits
+        self.map_path = None
+        self.render_iframe = False
         self.map_type = 'base'
+        self.map_id = '_'.join(['folium', uuid4().hex])
 
         #Mark counter, JSON, Plugins
         self.mark_cnt = {}
@@ -91,15 +132,18 @@ class Map(object):
         self.location = location
 
         #Map Size Parameters
+        self.width = width
+        self.height = height
         self.map_size = {'width': width, 'height': height}
         self._size = ('style="width: {0}px; height: {1}px"'
                       .format(width, height))
 
         #Templates
-        self.env = Environment(loader=PackageLoader('folium', 'templates'))
+        self.env = ENV
         self.template_vars = {'lat': location[0], 'lon': location[1],
                               'size': self._size, 'max_zoom': max_zoom,
-                              'zoom_level': zoom_start}
+                              'zoom_level': zoom_start,
+                              'map_id': self.map_id}
 
         #Tiles
         self.tiles = ''.join(tiles.lower().strip().split())
@@ -108,14 +152,16 @@ class Map(object):
                              ' or non-default Mapbox tiles.')
 
         self.default_tiles = ['openstreetmap', 'mapboxcontrolroom',
+                              'mapquestopen', 'mapquestopenaerial',
                               'mapboxbright', 'mapbox', 'cloudmade',
                               'stamenterrain', 'stamentoner']
         self.tile_types = {}
         for tile in self.default_tiles:
-            self.tile_types[tile] = {'templ':
-                                     self.env.get_template(tile + '_tiles.txt'),
-                                     'attr':
-                                     self.env.get_template(tile + '_att.txt')}
+            join_tile_path = functools.partial(os.path.join, 'tiles', tile)
+            self.tile_types[tile] = {
+                'templ': self.env.get_template(join_tile_path('tiles.txt')),
+                'attr': self.env.get_template(join_tile_path('attr.txt')),
+            }
 
         if self.tiles in self.tile_types:
             self.template_vars['Tiles'] = (self.tile_types[self.tiles]['templ']
@@ -127,10 +173,94 @@ class Map(object):
             if not attr:
                 raise ValueError('Custom tiles must'
                                  ' also be passed an attribution')
-            self.template_vars['attr'] = unicode(attr, 'utf8')
-            self.tile_types.update({'Custom': {'template': tiles, 'attr': attr}})
+            if isinstance(attr, binary_type):
+                attr = text_type(attr, 'utf8')
+            self.template_vars['attr'] = attr
+            self.tile_types.update({'Custom': {'template': tiles,
+                                               'attr': attr}})
 
-    def simple_marker(self, location=None, popup='Pop Text', popup_on=True):
+        self.added_layers = []
+        self.template_vars.setdefault('wms_layers', [])
+        self.template_vars.setdefault('tile_layers', [])
+
+    @iter_obj('simple')
+    def add_tile_layer(self, tile_name=None,tile_url=None,active=False):
+        '''adds a simple tile layer
+        Parameters
+        ----------
+        tile_name: string
+            name of the tile layer
+        tile_url: string
+            url location of the tile layer
+        active: boolean
+            should the layer be active when added
+        '''
+        if not tile_name in self.added_layers:
+            tile_name = tile_name.replace (" ", "_")
+            tile_temp = self.env.get_template('tile_layer.js')
+
+            tile = tile_temp.render({'tile_name': tile_name,
+                                       'tile_url': tile_url
+                                       })
+
+            self.template_vars.setdefault('tile_layers', []).append((tile))
+
+            self.added_layers.append({tile_name:tile_url})
+
+
+    @iter_obj('simple')
+    def add_wms_layer(self, wms_name=None, wms_url=None, wms_format=None,
+                      wms_layers=None, wms_transparent=True):
+        '''adds a simple tile layer
+        Parameters
+        ----------
+        wms_name: string
+            name of wms layer
+        wms_url : string
+            url of wms layer
+        '''
+        if not wms_name in self.added_layers:
+            wms_name = wms_name.replace (" ", "_")
+            wms_temp = self.env.get_template('wms_layer.js')
+
+            wms = wms_temp.render({
+                'wms_name': wms_name,
+                'wms_url': wms_url,
+                'wms_format': wms_format,
+                'wms_layer_names':wms_layers,
+                'wms_transparent':str(wms_transparent).lower()})
+
+            self.template_vars.setdefault('wms_layers', []).append((wms))
+
+            self.added_layers.append({wms_name:wms_url})
+
+    @iter_obj('simple')
+    def add_layers_to_map(self):
+        '''
+        Required function to actually add the layers to the html packet
+        '''
+        layers_temp = self.env.get_template('add_layers.js')
+
+        data_string = ''
+        for i, layer in enumerate(self.added_layers):
+            name = list(layer.keys())[0]
+            data_string+='\"'
+            data_string+=name
+            data_string+='\"'
+            data_string+=': '
+            data_string+=name
+            if i < len(self.added_layers)-1:
+                data_string+=",\n"
+            else:
+                data_string+="\n"
+
+        data_layers = layers_temp.render({'layers': data_string})
+        self.template_vars.setdefault('data_layers', []).append((data_string))
+
+    @iter_obj('simple')
+    def simple_marker(self, location=None, popup='Pop Text', popup_on=True,
+                      marker_color='blue', marker_icon='info-sign',
+                      clustered_marker=False, icon_angle=0):
         '''Create a simple stock Leaflet marker on the map, with optional
         popup text or Vincent visualization.
 
@@ -143,6 +273,13 @@ class Map(object):
             or a tuple of the form (Vincent object, 'vis_path.json')
         popup_on: boolean, default True
             Pass false for no popup information on the marker
+        marker_color
+            color of marker you want
+        marker_icon
+            icon from (http://getbootstrap.com/components/) you want on the marker
+        clustered_marker
+            boolean of whether or not you want the marker clustered with
+            other markers
 
         Returns
         -------
@@ -154,24 +291,89 @@ class Map(object):
         >>>map.simple_marker(location=[45.5, -122.3], popup=(vis, 'vis.json'))
 
         '''
-        self.mark_cnt['simple'] = count = self.mark_cnt.get('simple', 0) + 1
+        count = self.mark_cnt['simple']
 
         mark_temp = self.env.get_template('simple_marker.js')
 
+        marker_num = 'marker_{0}'.format(count)
+        add_line = "{'icon':"+marker_num+"_icon}"
+
+        icon_temp = self.env.get_template('simple_icon.js')
+        icon = icon_temp.render({'icon': marker_icon,
+                                 'icon_name': marker_num+"_icon",
+                                 'markerColor': marker_color,
+                                 'icon_angle': icon_angle
+                                })
+
         #Get marker and popup
         marker = mark_temp.render({'marker': 'marker_' + str(count),
-                                   'lat': location[0], 'lon': location[1]})
+                                   'lat': location[0],
+                                   'lon': location[1],
+                                   'icon': add_line
+                                   })
 
         popup_out = self._popup_render(popup=popup, mk_name='marker_',
                                        count=count,
                                        popup_on=popup_on)
 
-        add_mark = 'map.addLayer(marker_{0})'.format(count)
 
-        self.template_vars.setdefault('markers', []).append((marker,
+
+        if clustered_marker:
+            add_mark = 'clusteredmarkers.addLayer(marker_{0})'.format(count)
+            self.template_vars.setdefault('cluster_markers', []).append((icon,
+                                                             marker,
+                                                             popup_out,
+                                                             add_mark))
+        else:
+            add_mark = 'map.addLayer(marker_{0})'.format(count)
+            self.template_vars.setdefault('custom_markers', []).append((icon,
+                                                             marker,
                                                              popup_out,
                                                              add_mark))
 
+    @iter_obj('line')
+    def line(self, locations,
+             line_color=None, line_opacity=None, line_weight=None):
+        '''Add a line to the map with optional styles.
+
+        Parameters
+        ----------
+        locations: list of points (latitude, longitude)
+            Latitude and Longitude of line (Northing, Easting)
+        line_color: string, default Leaflet's default ('#03f')
+        line_opacity: float, default Leaflet's default (0.5)
+        line_weight: float, default Leaflet's default (5)
+
+        Note: If the optional styles are omitted, they will not be included
+        in the HTML output and will obtain the Leaflet defaults listed above.
+
+        Example
+        -------
+        >>>map.line(locations=[(45.5, -122.3), (42.3, -71.0)])
+        >>>map.line(locations=[(45.5, -122.3), (42.3, -71.0)],
+                    line_color='red', line_opacity=1.0)
+
+        '''
+        count = self.mark_cnt['line']
+
+        line_temp = self.env.get_template('polyline.js')
+
+        polyline_opts = {'color': line_color,
+                'weight': line_weight,
+                'opacity': line_opacity}
+
+        varname = 'line_{}'.format(count)
+        line_rendered = line_temp.render({'line': varname,
+                                          'locations': locations,
+                                          'options': polyline_opts})
+
+        add_line = 'map.addLayer({});'.format(varname)
+
+        self.template_vars.setdefault('lines', []).append((line_rendered,
+                                                           add_line))
+
+
+    @iter_obj('circle')
     def circle_marker(self, location=None, radius=500, popup='Pop Text',
                       popup_on=True, line_color='black', fill_color='black',
                       fill_opacity=0.6):
@@ -208,7 +410,7 @@ class Map(object):
                              radius=1000, popup=(bar_chart, 'bar_data.json'))
 
         '''
-        self.mark_cnt['circle'] = count = self.mark_cnt.get('circle', 0) + 1
+        count = self.mark_cnt['circle']
 
         circle_temp = self.env.get_template('circle_marker.js')
 
@@ -229,6 +431,7 @@ class Map(object):
                                                              popup_out,
                                                              add_mark))
 
+    @iter_obj('polygon')
     def polygon_marker(self, location=None, line_color='black', line_opacity=1,
                        line_weight=2, fill_color='blue', fill_opacity=1,
                        num_sides=4, rotation=0, radius=15, popup='Pop Text',
@@ -268,7 +471,7 @@ class Map(object):
 
         '''
 
-        self.mark_cnt['polygon'] = count = self.mark_cnt.get('polygon', 0) + 1
+        count = self.mark_cnt['polygon']
 
         poly_temp = self.env.get_template('poly_marker.js')
 
@@ -353,7 +556,7 @@ class Map(object):
             if isinstance(popup, str):
                 popup_temp = self.env.get_template('simple_popup.js')
                 return popup_temp.render({'pop_name': mk_name + str(count),
-                                          'pop_txt': popup})
+                                          'pop_txt': json.dumps(popup)})
             elif isinstance(popup, tuple):
                 #Update template with JS libs
                 vega_temp = self.env.get_template('vega_ref.txt').render()
@@ -370,7 +573,14 @@ class Map(object):
                 mark = ''.join([mk_name, str(count)])
                 json_out = popup[1]
                 div_id = popup[1].split('.')[0]
-                width, height = (vega.width + 75), (vega.height + 50)
+                width = vega.width
+                height = vega.height
+                if isinstance(vega.padding, dict):
+                    width += vega.padding['left']+vega.padding['right']
+                    height += vega.padding['top']+vega.padding['bottom']
+                else:
+                    width += 75
+                    height += 50
                 max_width = self.map_size['width']
                 vega_id = '#' + div_id
                 popup_temp = self.env.get_template('vega_marker.js')
@@ -380,8 +590,9 @@ class Map(object):
                                           'json_out': json_out,
                                           'vega_id': vega_id})
 
-    def geo_json(self, geo_path=None, data_out='data.json', data=None,
-                 columns=None, key_on=None, threshold_scale=None,
+    @iter_obj('geojson')
+    def geo_json(self, geo_path=None, geo_str=None, data_out='data.json',
+                 data=None, columns=None, key_on=None, threshold_scale=None,
                  fill_color='blue', fill_opacity=0.6, line_color='black',
                  line_weight=1, line_opacity=1, legend_name=None,
                  topojson=None, reset=False):
@@ -414,6 +625,8 @@ class Map(object):
         ----------
         geo_path: string, default None
             URL or File path to your GeoJSON data
+        geo_str: string, default None
+            String of GeoJSON, alternative to geo_path
         data_out: string, default 'data.json'
             Path to write Pandas DataFrame/Series to JSON if binding data
         data: Pandas DataFrame or Series, default None
@@ -456,19 +669,22 @@ class Map(object):
 
         Example
         -------
-        >>>map.geo_json(geo_path='us-states.json', line_color='blue', line_weight=3)
-        >>>map.geo_json(geo_path='geo.json', data=df, columns=['Data 1', 'Data 2'],
+        >>>map.geo_json(geo_path='us-states.json', line_color='blue',
+                        line_weight=3)
+        >>>map.geo_json(geo_path='geo.json', data=df,
+                        columns=['Data 1', 'Data 2'],
                         key_on='feature.properties.myvalue', fill_color='PuBu',
                         threshold_scale=[0, 20, 30, 40, 50, 60])
         >>>map.geo_json(geo_path='countries.json', topojson='objects.countries')
         '''
 
         if reset:
-            reset_vars = ['json_paths', 'func_vars', 'color_scales', 'geo_styles',
-                          'gjson_layers', 'map_legends', 'topo_convert']
+            reset_vars = ['json_paths', 'func_vars', 'color_scales',
+                          'geo_styles', 'gjson_layers', 'map_legends',
+                          'topo_convert']
             for var in reset_vars:
                 self.template_vars.update({var: []})
-            self.mark_cnt['geojson'] = 0
+            self.mark_cnt['geojson'] = 1
 
         def json_style(style_cnt, line_color, line_weight, line_opacity,
                        fill_color, fill_opacity, quant_fill):
@@ -486,11 +702,14 @@ class Map(object):
         #Set map type to geojson
         self.map_type = 'geojson'
 
-        #Set counter for GeoJSON and set iterations
-        self.mark_cnt['geojson'] = self.mark_cnt.get('geojson', 0) + 1
-
         #Get JSON map layer template pieces, convert TopoJSON if necessary
-        geo_path = ".defer(d3.json, '{0}')".format(geo_path)
+        # geo_str is really a hack
+        if geo_path:
+            geo_path = ".defer(d3.json, '{0}')".format(geo_path)
+        elif geo_str:
+            geo_path = (".defer(function(callback)"
+                        "{{callback(null, JSON.parse('{}'))}})").format(geo_str)
+
         if topojson is None:
             map_var = '_'.join(['gjson', str(self.mark_cnt['geojson'])])
             layer_var = map_var
@@ -512,6 +731,8 @@ class Map(object):
         #Get Data binding pieces if available
         if data is not None:
 
+            import pandas as pd
+
             #Create DataFrame with only the relevant columns
             if isinstance(data, pd.DataFrame):
                 data = pd.concat([data[columns[0]], data[columns[1]]], axis=1)
@@ -529,15 +750,18 @@ class Map(object):
 
             #D3 Color scale
             series = data[columns[1]]
+            if threshold_scale and len(threshold_scale) > 6:
+                raise ValueError
             domain = threshold_scale or utilities.split_six(series=series)
-            if len(domain) > 6:
-                raise ValueError('The threshold scale must be of length <= 6')
+            if len(domain) > 253:
+                raise ValueError('The threshold scale must be of length <= 253')
             if not utilities.color_brewer(fill_color):
                 raise ValueError('Please pass a valid color brewer code to '
                                  'fill_local. See docstring for valid codes.')
 
-            palette = utilities.color_brewer(fill_color)
+            palette = utilities.color_brewer(fill_color, len(domain))
             d3range = palette[0: len(domain) + 1]
+            tick_labels = utilities.legend_scaler(domain)
 
             color_temp = self.env.get_template('d3_threshold.js')
             d3scale = color_temp.render({'domain': domain,
@@ -548,6 +772,7 @@ class Map(object):
             name = legend_name or columns[1]
             leg_templ = self.env.get_template('d3_map_legend.js')
             legend = leg_templ.render({'lin_max': int(domain[-1]*1.1),
+                                       'tick_labels': tick_labels,
                                        'caption': name})
             self.template_vars.setdefault('map_legends', []).append(legend)
 
@@ -567,18 +792,23 @@ class Map(object):
         self.template_vars.setdefault('geo_styles', []).append(style)
         self.template_vars.setdefault('gjson_layers', []).append(layer)
 
-    def _build_map(self):
+    def _build_map(self, html_templ=None, templ_type='string'):
         '''Build HTML/JS/CSS from Templates given current map type'''
-        map_types = {'base': 'fol_template.html',
-                     'geojson': 'geojson_template.html'}
+        if html_templ is None:
+            map_types = {'base': 'fol_template.html',
+                         'geojson': 'geojson_template.html'}
 
-        #Check current map type
-        type_temp = map_types[self.map_type]
+            #Check current map type
+            type_temp = map_types[self.map_type]
+
+            html_templ = self.env.get_template(type_temp)
+        else:
+            if templ_type == 'string':
+                html_templ = self.env.from_string(html_templ)
 
-        html_templ = self.env.get_template(type_temp)
         self.HTML = html_templ.render(self.template_vars)
 
-    def create_map(self, path='map.html', plugin_data_out=True):
+    def create_map(self, path='map.html', plugin_data_out=True, template=None):
         '''Write Map output to HTML and data output to JSON if available
 
         Parameters:
@@ -588,20 +818,53 @@ class Map(object):
         plugin_data_out: boolean, default True
             If using plugins such as awesome markers, write all plugin
             data such as JS/CSS/images to path
+        template: string, default None
+            Custom template to render
 
         '''
+        self.map_path = path
+        self._build_map(template)
 
-        self._build_map()
-
-        with codecs.open(path, 'w', 'utf-8') as f:
+        with codecs.open(path, 'w', 'utf8') as f:
             f.write(self.HTML)
 
         if self.json_data:
-            for path, data in self.json_data.iteritems():
+            for path, data in iteritems(self.json_data):
                 with open(path, 'w') as g:
                     json.dump(data, g)
 
         if self.plugins and plugin_data_out:
-            for name, plugin in self.plugins.iteritems():
+            for name, plugin in iteritems(self.plugins):
                 with open(name, 'w') as f:
+                    if isinstance(plugin, binary_type):
+                        plugin = text_type(plugin, 'utf8')
                     f.write(plugin)
+
+    def _repr_html_(self):
+        """Build the HTML representation for IPython."""
+        map_types = {'base': 'ipynb_repr.html',
+                     'geojson': 'ipynb_iframe.html'}
+
+        #Check current map type
+        type_temp = map_types[self.map_type]
+        if self.render_iframe:
+            type_temp = 'ipynb_iframe.html'
+        templ = self.env.get_template(type_temp)
+        self._build_map(html_templ=templ, templ_type='temp')
+        if self.map_type == 'geojson' or self.render_iframe:
+            if not self.map_path:
+                raise ValueError('Use create_map to set the path!')
+            return templ.render(path=self.map_path, width=self.width,
+                                height=self.height)
+        return self.HTML
+
+    def display(self):
+        """Display the visualization inline in the IPython notebook.
+
+        This is deprecated, use the following instead::
+
+            from IPython.display import display
+            display(viz)
+        """
+        from IPython.core.display import display, HTML
+        display(HTML(self._repr_html_()))
